package require Iocpsock
rename socket socketo
rename socket2 socket



proc doc {} {
  # out tuplespace tuple args
  #  write tuple to tuplespace
  # dir tuplespace ?pattern?
  #  list the tuples in tuplespace matching ?pattern? if pattern is not provided list all tuples
  # dir
  #  list all tuplespaces
  # open tuplespace
  #  create tuplespace if not already exists
  # close tuplespace
  #  close tuplespace and remove all remaining tuples in this space
  # exit
  #  disconnect session
  # read space tuplemask
  #  read a tuple matching tuplemask from tuplespace. The tuple will remain in the space
  # in space tuplemask
  #  read and remove a tuple matching tuplemask from tuplespace
  # ain
  # aread
  #  asynchronous versions of read and in
  # ping
  #  check connection with server
  # help
  #  returns all available server commands
}

proc init {} {
  # open persistent database
  package require Tk
  package require sqlite3

  # sqlite3 db ~/ts.sqlite
  sqlite3 db {}
  db eval {PRAGMA synchronous=OFF;}

  set server_port 4040
  while { [catch {socket -server handle_socket $server_port}] } {after 1000}
  label .l -text "Server listening on port $server_port, F5 for console , F6 for restart"
  pack .l
  bind . <F5> {console show}
  bind . <F6> {exec wish $argv0 &; exit}

  # aliases that can be used from console
  interp alias {} ts_out {} ::ts::out stdout
  interp alias {} ts_ain {} ::ts::ain stdout
  interp alias {} ts_in {} ::ts::in stdout
  interp alias {} ts_aread {} ::ts::aread stdout
  interp alias {} ts_read {} ::ts::read stdout
  interp alias {} ts_dir {} ::ts::dir stdout
  interp alias {} ts_open {} ::ts::open stdout
  interp alias {} ts_close {} ::ts::close stdout
}

proc handle_socket {socket host port} {
  puts "Incoming connection from $host ($socket)"
  fconfigure $socket -buffering line -blocking no
  fileevent $socket readable [list ::ts::handle_request $socket]
}

# callback namespace
namespace eval cb {
}

# namespace for server commands
namespace eval ts {

  # commands supported by the tupleserver
  set commands {open close dir ain in aread read out exit ping help}

  # checks if table names that are sent to Sqlite are sane to prevent SQL injection
  # throws an error if they are not
  proc issane {name} {
    set sname [regsub -all {[^[:alnum:]]} $name {}]
    if {($name ne $sname) || ($name eq {})} {
      error "invalid tuplespace name: \"$name\""
    }
  }

  # check if argument is a list. TODO replace with [string is list] in 8.5 beta
  proc islist {string} {
    # check if a string is a valid list before doing list ops
    if {[catch {lindex $string 0}]} {
      return false
    } else {
      return true
    }
  }

  # match a tuple with a pattern
  proc tuple_match {tuple pattern} {
    if {[llength $tuple] ne [llength $pattern]} {
      return false
    }
    foreach tup $tuple pat $pattern {
      if {![string match $pat $tup]} {
        return false
      }
    }
    return true
  }

  # shuffle proc for load balancing callbacks
  proc shuffle { list } {
    set len [llength $list]
    set len2 $len
    for {set i 0} {$i < $len-1} {incr i} {
      set n [expr {int($i + $len2 * rand())}]
      incr len2 -1
      set temp [lindex $list $i]
      lset list $i [lindex $list $n]
      lset list $n $temp
    }
    return $list
  }

  # handle incoming request on the tupleserver
  proc handle_request {socket} {
    variable commands
    if {[catch {gets $socket req}] } return
    if {[chan eof $socket]} {
      puts "connection with $socket closed"
      chan close $socket
      foreach var [info vars ::cb::*] {
        unset -nocomplain ${var}($socket)
      }
      return
    }
    if {$req eq {}} {
      # not a full line yet
      return
    }
    if {![islist $req]} {
      catch {puts $socket [list error "request should be a valid Tcl list"]}
      return
    }
    set cmd [lindex $req 0]
    set params [lrange $req 1 end]
    puts "Handling request: $cmd, params $params"
    if {$cmd eq {}} { set cmd ping }
    if {[lsearch -all $commands $cmd] eq {}} {
      catch {puts $socket [list error "unknown server command $cmd"]}
      return
    }
    set status [catch {$cmd $socket {*}$params} result]
    puts "$socket $status: $result"
    switch $status {
      2 {return}
      1 {
        catch {puts $socket [list error $result]}
      }
      0 {
        catch {puts $socket [list ok $result]}
      }
    }

  }

  proc ping {_ args } {
    return
  }
  proc help {_ args} {
    variable commands
    return $commands
  }

  proc dir {_ args } {
    if {([llength $args] > 2)} {
      error "wrong # of args, should be \"dir ?tuplespace? ?pattern?\""
    }
    set tuplespace [lindex $args 0]
    set pattern [lindex $args 1]
    if {$tuplespace eq {}} {
      set tuplespaces {}
      db eval {SELECT * from sqlite_master WHERE type='table' AND name NOT LIKE 'sqlite%'} {
        lappend tuplespaces $name
      }
      return $tuplespaces
    } else {
      issane $tuplespace
      if {[lsearch [dir $_] $tuplespace] < 0} {
        error "unknown tuplespace \"$tuplespace\""
      }
      set tuples {}
      db eval [format {SELECT tuple from "%s"} $tuplespace] {
        if {$pattern ne {}} {
          if {![tuple_match $tuple $pattern]} {
            continue
          }
        }
        lappend tuples $tuple
      }
      return $tuples
    }
  }

  proc open {_ args} {
    if {[llength $args] != 1} {
      error "wrong # of args, should be \"open tuplespace\""
    }
    set tuplespace [lindex $args 0]
    issane $tuplespace
    db eval [format {CREATE TABLE IF NOT EXISTS "%s"(id integer primary key autoincrement, tuple string)} $tuplespace]
    return
  }

  proc close {_ args} {
    if {[llength $args] != 1} {
      error "wrong # of args, should be \"close tuplespace\""
    }
    set tuplespace [lindex $args 0]
    issane $tuplespace
    # close any pending callbacks for this tuplespace
    foreach socket [array names ::cb::${tuplespace}] {
      catch {puts $socket [list error "tuplespace $tuplespace destroyed"]}
    }
    array unset ::cb::${tuplespace}
    return [db eval [format {DROP TABLE IF EXISTS "%s"} $tuplespace]]
  }

  proc out {_ args} {
    if {[llength $args] != 2} {
      error "wrong # of args, should be \"out tuplespace tuple\""
    }
    set tuplespace [lindex $args 0]
    set tuple [lindex $args 1]
    issane $tuplespace

    foreach socket [shuffle [array names ::cb::${tuplespace}]] {
      lassign [set ::cb::${tuplespace}($socket)] type pattern
      if {[tuple_match $tuple $pattern]} {
        catch {puts $socket [list ok $tuple]}
        unset ::cb::${tuplespace}($socket)
        if {$type eq "in"} {
          return
        }
      }
    }
    return [db eval [format {INSERT INTO "%s" VALUES(NULL,$tuple)} $tuplespace]]
  }

  proc aread {socket args} {
    if {[llength $args] != 2} {
      error "wrong # of args, should be \"aread tuplespace pattern\""
    }
    set tuplespace [lindex $args 0]
    set pattern [lindex $args 1]
    issane $tuplespace
    db eval [format {SELECT tuple FROM "%s"} $tuplespace] {
      if {[tuple_match $tuple $pattern]} {
        return $tuple
      }
    }
    return {}
  }

  proc read {socket args} {
    if {[llength $args] != 2} {
      error "wrong # of args, should be \"read tuplespace pattern\""
    }
    set tuplespace [lindex $args 0]
    set pattern [lindex $args 1]
    issane $tuplespace
    set resp [aread $socket $tuplespace $pattern]
    if {$resp eq {}} {
      set ::cb::${tuplespace}($socket) [list read $pattern]
      return -code return
    } {
      return $resp
    }
  }

  proc ain {_ args} {
    if {[llength $args] != 2} {
      error "wrong # of args, should be \"ain tuplespace pattern\""
    }
    set tuplespace [lindex $args 0]
    set pattern [lindex $args 1]
    issane $tuplespace
    set id {}
    db transaction {
      db eval [format {SELECT id,tuple FROM "%s"} $tuplespace] {
        if {[tuple_match $tuple $pattern]} {
          break
        } else {
          set id {}
        }
      }
      if {$id ne {}} {
        db eval [format {DELETE FROM "%s" WHERE id='%s'} $tuplespace $id]
        return $tuple
      } {
        return
      }
    }
  }

  proc in {socket args} {
    if {[llength $args] != 2} {
      error "wrong # of args, should be \"in tuplespace pattern\""
    }
    set tuplespace [lindex $args 0]
    set pattern [lindex $args 1]
    issane $tuplespace
    set resp [ain $socket $tuplespace $pattern]
    if {$resp eq {}} {
      set ::cb::${tuplespace}($socket) [list in $pattern]
      return -code return
    } {
      return $resp
    }
  }

  proc exit {socket} {
    chan close $socket
    return -code return
  }
}

init
vwait forever
